{- 
 - IRC-to-Rage-Comic app
 - by Vidar 'koala_man' Holen
 - http://www.vidarholen.net/contents/rage
 -
 - Copyright or something I guess, 2011
 -
 - If you use it for anything, I'd appreciate a 
 - small notice mentioning your source of inspiration. 
 - Also maybe an email to vidar@vidarholen.net so I can try it!
 -
 - This app will take IRC log as input, attempt to parse it,
 - and spit out some partial html that with the proper 
 - wrapping and style sheet could look like a rage comic. 
 -
 - Compile with ghc --make rage.hs
 -
 -}
import Text.Regex.TDFA
import Control.Monad
import Data.Maybe
import Data.List
import Data.Char 
import System.Random
import Text.Regex

type Face = String
type Html = String

data Direction = ToLeft | ToRight deriving (Show)

data Comment = Comment { 
    nick :: String,
    text :: [String],
    face :: Face
} deriving (Show)

data Panel = Panel [Comment] deriving (Show)
data Comic = Comic [Panel] deriving (Show)

data RawComment = RawComment String String deriving (Show)

--------------------------------------------------------

toLowercase = map toLower

divideInto _ [] = []
divideInto n list = 
    let (first, rest) = splitAt n list
    in
        first : (divideInto n rest)

--------------------------------------------------------

getGroups s re = let (_, _, _, groups) = (s =~ re) :: (String,String,String,[String]) in 
    if null groups then Nothing else Just groups

normalParse s = do 
    substrings <- getGroups s "^ *([0-9:]* *)?<[ ]?([^>]*)> (.*) *$" 
    return $ RawComment (substrings !! 1) (substrings !! 2)

normalActionParse s = do 
    substrings <- getGroups s "^ *([0-9:]* *)?\\* +([^ ]+) (.*) *$" 
    return $ RawComment (substrings !! 1) $ "*" ++ (substrings !! 2) ++ "*"

desperateAngleBracketParse s = do
    substrings <- getGroups s "^[^<]*<([^>]+)> *(.*)$" 
    return $ RawComment (substrings !! 0) (substrings !! 1)

halfAngleBracketParse s = do
    substrings <- getGroups s "^([ \t]*[][()0-9:.]*[ \t]+)?([^<>]+)> *(.*)$" 
    return $ RawComment (substrings !! 1) (substrings !! 2)

desperateParenthesisParse s = do
    substrings <- getGroups s "^[^(]*\\(([^)]+)\\) *(.*)$" 
    return $ RawComment (substrings !! 0) (substrings !! 1)

desperateSquareBracketParse s = do
    substrings <- getGroups s "^[^[]*\\[([^]]+)] *(.*)$" 
    return $ RawComment (substrings !! 0) (substrings !! 1)
    
colonParse s = do
    substrings <- getGroups s "^ *([^:]{0,15}): (.*) *$"
    return $ RawComment (substrings !! 0) (substrings !! 1)

parseLine s = foldl mplus Nothing $ map ($s) [normalParse, normalActionParse, halfAngleBracketParse, desperateAngleBracketParse, desperateParenthesisParse, desperateSquareBracketParse, colonParse]

parse s = catMaybes . map parseLine . lines $ s

rawToFull (RawComment nick text) = Comment nick [text] "happy" 

--------------------------------------------------------



makeComic comments = Comic $ map Panel $ divideInto 2 $ collateByNick comments

collateByNick :: [Comment] -> [Comment]
collateByNick (f:r) = collate f 0 r
    where
        collate l _ [] = [l]
        collate current n (next:rest) = 
            if (nick current == nick next) && n < 3 then 
                let (Comment nick list face) = current 
                    new = Comment nick (list ++ (text next)) face 
                in
                    collate new (n+1) rest 
            else
                current : collate next 0 rest


data Emotion = Normal | Lol | Happy | Sad | Angry | Kitteh deriving (Show, Eq)

endFaces = [ "epicwin", "clean", "kittehsmile", "monocole", "redeyes", "french" ]

normalFaces = [ "happy", "beh", "concentrated", "dude-come-on", "herp", "hehehe", "smile", "suspicious", "wait", "hehehe" ]
kittehFaces = ["kittehsmile" ]
happyFaces = [ "epicwin", "grin", "kittehsmile" ]
lolFaces = [ "pfftch", "loool" ]
angryFaces = [ "redeyes", "super-rage"]
sadFaces = [ "unhappy", "whyyyyyy", "herp", "horror" ] 

classify text = fromJust $ 
    Nothing 
    `mplus` match "<3"       Kitteh 
    `mplus` match "\\^ *\\. *\\^|\\^_+\\^" Kitteh 
    `mplus` match ":-?[])DPp]"       Happy 
    `mplus` match ":'*-?[[(]"         Sad    
    `mplus` match "^:-?/|[^p]:-?[/]"         Sad    
    `mplus` match ";_+;"         Sad    
    `mplus` matchL "lo+l|(h[eioa]h[eioa])|rofl|lmao" Lol   
    `mplus` matchL "mjau|meow|katt|kitte[nh]ponni|pony|bunny|kanin|marsvin|nuss|klem" Kitteh 
    `mplus` matchL "fff"        Angry  
    `mplus` match "![!1]"        Angry  
    `mplus` matchL "faen|satan|helvete|dritt|fuck[^i]|damn" Angry


    `mplus` Just Normal
    where 
        textLow = toLowercase (text ++ " ") 
        textAsIs = text ++ " "
        matchL s v = if textLow  =~ s then Just v else Nothing
        match s v =  if textAsIs =~ s then Just v else Nothing


attachEmotions = map (\x -> (classify . unlines $ text x, x)) 

collateByNickAndEmotion [] = []
collateByNickAndEmotion ((fe, fc):r) = collate fe fc 0 r
    where
        collate e l _ [] = [(e,l)]
        collate currentEmotion currentComment n ((nextE, next):rest) = 
            if (nick currentComment == nick next && (currentEmotion == nextE || currentEmotion == Normal || nextE == Normal )) && n < 4 then 
                let (Comment nick list face) = currentComment
                    new = Comment nick (list ++ (text next)) face 
                in
                    collate (maxOf nextE currentEmotion) new (n+1) rest 
            else
                (currentEmotion, currentComment) : collate nextE next 0 rest
        
        maxOf a Normal = a
        maxOf Normal a = a
        maxOf a _ = a

getFacesFor emotion = case emotion of 
                        Normal -> normalFaces
                        Kitteh -> kittehFaces
                        Sad -> sadFaces
                        Angry -> angryFaces
                        Happy -> happyFaces
                        Lol -> lolFaces

getFaceIn list randomNumber = 
        let len = length list in
            list !! (randomNumber `mod` len)

getFaceFor emotion randomNumber = 
    let list = getFacesFor emotion
    in 
        getFaceIn list randomNumber
            
assignFace (e, Comment nick text _) = do
    n <- newStdGen 
    let (r,g) = randomR (0,100000) n 
    return $ Comment nick text (getFaceFor e r)


massageCommentList comments = do
    let ec = attachEmotions comments
    let collated = collateByNickAndEmotion ec
    assigned <- mapM assignFace collated 
    let lastemotion = if null collated then Normal else fst $ last collated
    if lastemotion == Normal 
        then do
            f <- addZinger assigned
            return f
        else do
            return assigned

addZinger :: [Comment] -> IO [Comment]
addZinger [] = return []
addZinger ((Comment nick text _):[])  = do 
    n <- newStdGen :: IO StdGen
    let (r,_) =  randomR (0,1000000) n
    return $ (Comment nick text (getFaceIn endFaces $ r `div` 1000)) : []
    
addZinger (f:r) = do
    rest <- addZinger r
    return $ f:rest

makePanels [] = []
makePanels (first:rest) = mp first rest
    where mp c [] = Panel [c] : []
          mp c (next:rest) = 
            if nick c == nick next 
            then Panel [c] : (mp next rest) 
            else Panel [c,next] : makePanels rest

--------------------------------------------------------

commentToDiv (Comment nick lines face) width lr = 
    "<div class='comment comment"++(show lr)++" sharedWith" ++ (show width) ++ "'>" 
    ++ commentsTag lines 
    ++ faceTag face lr
    ++ "<br /><div class='nick'>" ++ nick ++"</div>"
    ++ "</div>"

addLinks s = rep "(http://([^ /]*)[^ ]*)" "<a href='\\1'>\\2/...</a>" s
    where rep regex replacement source = subRegex (mkRegex regex) source replacement

escapeHtml s = 
      rep "<" "&lt;" $ rep ">" "&gt;" $ rep "&" "&amp;" $ s
    where rep regex replacement source = subRegex (mkRegex regex) source replacement

commentTag comment = "<span class='line'>" ++ (addLinks $ escapeHtml comment) ++ "</span>\n"
commentsTag comments = 
    "<div class='comments'>\n" 
    ++ (intercalate "<br /><b>&nbsp;&nbsp;&nbsp;|</b><br />" $ map commentTag comments) 
    ++ "<br />&nbsp;&nbsp;&nbsp;&nbsp;<b>\\</b><br />"
    ++ "</div>\n"

faceTag face lr = "<img src=\"img/" ++ face ++ ".png\" class='face face" ++ (show lr) ++ "' alt='"++face++"' />\n"

panelToDiv (Panel comments) = 
    "<div class='panel'>\n" 
    ++ (commentToDiv (head comments) width ToRight) 
    ++ (concat $ map (\x -> commentToDiv x width ToLeft) $ tail comments) 
    ++ "</div>\n"
    where width = length comments

addTags :: String -> [String] -> String
addTags tag strings = "<"++tag++">" ++ (intercalate ("</"++tag++">\n<"++tag++">") strings) ++ "</"++tag++">\n"

comicToDiv (Comic panels) = 
    let pairs = divideInto 2 panels
        stringPairs = map (map panelToDiv) pairs 
        cells = map (addTags "td") stringPairs 
        rows = addTags "tr" cells 
    in
        "<table class='comic'>\n" ++ rows ++ "</table>\n"


comicToHtml c = htmlHeader ++ comicToDiv c

htmlHeader = ""


--------------------------------------------------------



main = do
    input <- getContents
    putStrLn "<!-- Original rage comic generator thing by Vidar 'koala_man' Holen -->"
    putStrLn $ "<div style=\"display: none; visibility: hidden\">\n This is a log to help parsing in the future: \n" ++ (escapeHtml input) ++ "\n</div>\n" 
    let parsed = map rawToFull . parse $ input
    if null parsed 
        then do
            putStrLn "I couldn't parse any of that :OOO<br />"
            putStrLn "I can leniently handle a bunch of formats, but apparently not yours at the moment. Feel free to try again with a format closer to the one in the example! "
        else do
            preprocessed <- massageCommentList parsed 
            putStrLn $ comicToHtml . Comic .  makePanels $ preprocessed


